/*
 * Copyright 2005,2006 Red Hat, Inc.
 *
 * Author: Chris Feist <cfeist@redhat.com>
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rpcsvc/ypclnt.h>
#include "automount.h"

#define MODPREFIX "nsswitch: "

#define MAPTYPE_FILE 1
#define MAPTYPE_PROGRAM 2

int underscore_to_dot = 0;

static int isfilemap(const char *);
static int isypmap(const char *);

char *make_mapname(const char *loc, const char *type)
{
	int retsize, needs_etc = 0;
	char *retval;

	retsize = strlen(loc) + strlen(type) + 2;
       	if (strncmp(loc, "/etc/", 5)) {
		retsize += 5;
		needs_etc = 1;
	}

	retval = malloc(retsize);
	if (!retval)
		return NULL;

	snprintf(retval, retsize, "%s:%s%s", type,
		 needs_etc ? "/etc/" : "",loc);

	return retval;
}

/*
 * Function which takes in a partial map name (ie. auto.misc), parses the
 * nsswitch.conf file and returns a valid map name (ie. yp:auto.misc or
 * file:/etc/auto.misc).  "cur" is the current map we are reading.  This
 * is used when determining the source of an included map.  For example.
 * if we are parsing /etc/auto.master, and it has the line "+auto.master",
 * we need to grab the auto.master from nis, if nis is specified in
 * nsswitch.conf.
 */
char *get_nsswitch_map(const char *loc, char *cur)
{
	char buf[1024];
	char *ordering;
	const char *automount_str = "automount:";
	char *comment = NULL;
	FILE *nsswitch;
	int found_automount = 0;
	char *retval = NULL;
	int retsize = 0;
	char *ypmapname = (char *)loc;

	debug(MODPREFIX "called nsswitch with: '%s' '%s'", loc, cur);
	nsswitch = fopen(_PATH_NSSWITCH_CONF, "r");
	if (!nsswitch) {
		error(MODPREFIX "Unable to open %s", _PATH_NSSWITCH_CONF);
		return NULL;
	}

	while ((ordering = fgets((char *)buf, sizeof(buf), nsswitch))) {
		if ((comment = strchr(ordering,'#')))
			*comment = '\0';
		while (isspace(*ordering)) ordering++;
		if (!strncmp(ordering, automount_str, sizeof(automount_str))) {
			ordering += strlen(automount_str);
			found_automount = 1;
			break;
		}
	}

	fclose(nsswitch);

	if (!found_automount)
		return NULL;

	/*
	 *  We support an option to convert underscores to dots when looking
	 *  up yp maps.
	 */
	if (underscore_to_dot) {
		char *u;

		ypmapname = strdup(loc);
		if (!ypmapname)
			return NULL;
		while ((u = strchr(ypmapname, '_')))
			*u = '.';
	}

	while (*ordering != '\0') {
		while (isspace(*ordering) && (*ordering != '\0')) ordering++;
		if (!strncmp(ordering, "files", 5)) {
			switch (isfilemap(loc)) {
			case MAPTYPE_FILE:
				retval = make_mapname(loc, "file");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			case MAPTYPE_PROGRAM:
				retval = make_mapname(loc, "program");
				debug("comparing %s to %s\n", retval+5,cur);
				if (cur && !strcmp(cur, retval+5)) {
					free(retval);
					retval = NULL;
					break;
				}
				return retval;
			default: // filemap doesn't exist
				break;
			}

		} else if ((!strncmp(ordering, "yp", 2) ||
			    !strncmp(ordering,"nis", 3)) &&
			   isypmap(ypmapname)) {
			retsize = strlen(ypmapname) + 4;
			retval = malloc(retsize);
			snprintf(retval, retsize, "yp:%s", ypmapname);
			return retval;
		}
		while (!isspace(*ordering) && (*ordering != '\0')) ordering++;
	}
	error(MODPREFIX "couldn't find map %s", loc);
	return retval;
}

/*
 * Function takes in a filename and tests if it exists in "/etc/"
 * Returns: MAPTYPE_FILE if it is not executable, MAPTYPE_PROGRAM if it
 * is executable and 0 if it doesn't exists or has incorrect permissions.
 */

static int isfilemap(const char *loc)
{
	struct stat st;
	int ret = 0;	
	char *realfilemap;

	if (*loc != '/') {
		realfilemap = malloc(strlen(loc) + 6); /* '/etc/' + '\0' */
		if (!realfilemap) {
			crit(MODPREFIX "malloc failed.");
			return 0;
		}
		snprintf(realfilemap, strlen(loc) + 6, "/etc/%s", loc);
		ret = stat(realfilemap, &st);
		free(realfilemap);
	} else
		ret = stat(loc, &st);

	if (!ret) {
		if (st.st_uid != 0) {
			error(MODPREFIX "file /etc/%s exists but is not"
					" owned by root.", loc);
			return 0;
		} else if (st.st_mode & S_IRUSR) {
			if (st.st_mode & S_IXUSR) 
				return MAPTYPE_PROGRAM;
			else
				return MAPTYPE_FILE;
		}
	}
	debug(MODPREFIX "stat(%s) returned errno %d\n", loc, errno);
	return 0;
}

/*
 * Function takes in a yp map name and returns
 * 1 if it exists or 0 if it doesn't.
 *
 * Some of this code borrowed from ypcat
 */

static int isypmap(const char *loc)
{
	int err;
	char *domainname;
	unsigned int order;

	if ((err = yp_get_default_domain(&domainname)) != YPERR_SUCCESS) {
		error(MODPREFIX "unable to get default yp domain");
		return 0;
	}
	if ((err = yp_order(domainname, loc, &order)) != YPERR_SUCCESS) {
		debug(MODPREFIX "unable to find map, %s in domain, %s",
		      loc, domainname);
		return 0;
	}

	return 1;
}

#ifdef STANDALONE
int do_debug = 0;

void print_usage(const char *prog)
{
	fprintf(stderr, "Usage %s [-u] <mapname>\n", basename(prog));
}

int main(int argc, char **argv)
{
	char *map;
	int  opt;

	if (argc < 2) {
		print_usage(argv[0]);
		exit(1);
	}

	while ((opt = getopt(argc, argv, "u")) != -1) {
		switch (opt) {
		case 'u':
			underscore_to_dot = 1;
			break;
		case '?':
		default:
			print_usage(argv[0]);
			exit(1);

		}
	}

	map = get_nsswitch_map(argv[optind], (char *)"auto.master");
	if (!map)
		exit(2);

	fprintf(stdout, "%s\n", map);
	free(map);
	exit(0);
}
#endif
