#ident "$Id: lookup_file.c,v 1.4 2004/01/29 16:01:22 raven Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *  lookup_file.c - module for Linux automount to query a flat file map
 *
 *   Copyright 1997 Transmeta Corporation - All Rights Reserved
 *   Copyright 2001-2003 Ian Kent <raven@themaw.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */


#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>
#include <signal.h>
#include <dlfcn.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rpcsvc/ypclnt.h>

#define MODULE_LOOKUP
#include "automount.h"

#define MAPFMT_DEFAULT "sun"

#define MAPTYPE_FILE	1
#define MAPTYPE_YP	2

#define MODPREFIX "lookup(file): "

typedef enum {
	st_begin, st_compare, st_star, st_badent, st_entspc, st_getent
} LOOKUP_STATE;

typedef enum { got_nothing, got_star, got_real } FOUND_STATE;
typedef enum { esc_none, esc_char, esc_val } ESCAPES;

/*
 *  This is the basic data structure used to store information about a map.
 *
 *  next:  linked list pointer pointing at the next map in the list.  There
 *         will only be multiple entries in this list if there are included
 *         maps (map entries of the form "+mapname").
 *  name:  This is the expanded name of the map.  For example, given a map
 *         entry "+auto.foo", this would be either "auto.foo" if it is resolved
 *         to a yp map, or "/etc/auto.foo" if it is a file map.  The type of
 *         map is determined by consulting the nsswitch.conf file.  The
 *         expanded map name is stored so that we don't have to lookup the
 *         file system path name more than once.  For an example of where
 *         this is useful, see the map_modified function.
 *  mapentry: If this is an included map, then this field represents the
 *         name specified in the "+auto.xyz" entry.  It can be the same as
 *         the "name" entry (annotated above) if the included map is a yp
 *         map or specifies an absolute path.
 *  type:  Either MAPTYPE_FILE (which includes both file maps and program
 *         maps), or MAPTYPE_YP.  MAPTYPE_LDAP is not yet implemented.
 *  u:     This union contains the maptype specific modification time.  For
 *         file maps, this is simply the mtime.  For yp maps, we use the yp
 *         order, or version number.
 *
 *  When adding a map to the list, the name, type, and mapentry must be
 *  specified.
 */
struct map {
	struct map *next;
	char *name;
	char *mapentry;
	int type;
	union {
		time_t mtime;
		unsigned int order;
	} u;
};

struct lookup_context {
	const char *mapname;
	struct parse_mod *parse;
	struct map *maps; /* linked list of maps associated with ctxt */

	/* The following are used for included yp maps */
	void *yplib_dh; /* handle returned from dlopen */
	lookup_one_t lookup_one_included;
	lookup_readmap_t lookup_readmap_included;
};

static int read_map(const char *root, time_t now, struct lookup_context *ctxt);
static int __read_map(char *, const char *,
		      time_t, struct lookup_context *);
static int __lookup_one(const char *map, const char *root, const char *key,
			int key_len, struct lookup_context *ctxt);
static int lookup_one(const char *root, const char *key, int key_len,
		      struct lookup_context *ctxt);
static int add_map(struct lookup_context *ctxt, int type, const char *mapname,
		   const char *mapentry);
unsigned int map_order(const char *mapname);

int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */

int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
{
	struct lookup_context *ctxt;
	struct stat st;

	if (!(*context = ctxt = malloc(sizeof(struct lookup_context)))) {
		crit(MODPREFIX "malloc: %m");
		return 1;
	}
	memset(ctxt, 0, sizeof(*ctxt));

	if (argc < 1) {
		crit(MODPREFIX "No map name");
		return 1;
	}

	ctxt->mapname = argv[0];
	if (ctxt->mapname[0] != '/') {
		crit(MODPREFIX "file map %s is not an absolute pathname",
		       ctxt->mapname);
		return 1;
	}

	if (access(ctxt->mapname, R_OK)) {
		crit(MODPREFIX "file map %s missing or not readable",
		       ctxt->mapname);
		return 1;
	}

	if (stat(ctxt->mapname, &st)) {
		crit(MODPREFIX "could not stat file map: %s",  ctxt->mapname);
		return 1;
	}

	if (!mapfmt)
		mapfmt = MAPFMT_DEFAULT;

	cache_init();

	return !(ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1));
}

static void recursion_detected(struct lookup_context *ctxt, const char *mapname)
{
	struct map *cur = ctxt->maps;

	crit(MODPREFIX "Recursion in included maps detected.\n");
	for (cur = ctxt->maps; cur; cur = cur->next)
		crit(MODPREFIX "%s", cur->name);
	crit(MODPREFIX "%s\n", mapname);
}

static void map_list_init(struct lookup_context *ctxt)
{
	struct map *cur, *next;

	cur = ctxt->maps;
	ctxt->maps = NULL;

	while (cur) {
		next = cur->next;
		free(cur->name);
		free(cur->mapentry);
		free(cur);
		cur = next;
	}
}

/*
 *  Add this map to the list of maps associated with this context.  Check
 *  for recursion in included maps, here.
 *
 *  Returns 1 on success, 0 on failure.
 */
static int add_map(struct lookup_context *ctxt, int type, const char *mapname,
	const char *mapentry)
{
	struct map *map, *cur, *tail;

	debug(MODPREFIX "Adding map %s,%s\n", mapname, mapentry);
	if (type < MAPTYPE_FILE || type > MAPTYPE_YP)
		return 0;

	map = malloc(sizeof(*map));
	if (!map) {
		crit("malloc of size %d failed while trying to read map %s."
		     " Continuing, but map contents may not be complete.\n",
		     sizeof(*map), mapname);
		return 0;
	}
	memset(map, 0, sizeof(*map));

	map->type = type;
	map->name = strdup(mapname);
	map->mapentry = strdup(mapentry);
	if (!map->name || !map->mapentry) {
		crit("memory allocation failed while trying read map %s."
		     "  Continuing, but map contents may not be complete.\n",
		     mapname);
		free(map->name);
		free(map->mapentry);
		free(map);
		return 0;
	}

	/* check for recursive includes */
	for (cur = tail = ctxt->maps; cur; cur = cur->next) {
		if (!strcmp(cur->name, mapname)) {
			recursion_detected(ctxt, mapname);
			free(map->name);
			free(map);
			return 0;
		}
		tail = cur;
	}
	if (tail)
		tail->next = map;
	else
		ctxt->maps = map;

	debug("added map name=\"%s\", entry=\"%s\" to context\n", map->name,
	      map->mapentry);
	return 1;
}

/*
 *  Remove this map from the list of maps associated with this context.
 *
 *  Returns 1 on success, 0 on failure.
 */
static int del_map(struct lookup_context *ctxt, const char *mapname)
{
	struct map *cur, *prev = ctxt->maps;

	for (cur = ctxt->maps; cur; cur = cur->next) {
		if (strcmp(cur->name, mapname)) {
			prev->next = cur->next;
			free(cur->name);
			free(cur);
			return 1;
		}
	}

	return 0;
}

static struct map *find_map_byname(struct lookup_context *ctxt, char *mapname)
{
	struct map *cur;

	for (cur = ctxt->maps; cur; cur = cur->next) {
		if (!strcmp(cur->name, mapname))
			return cur;
	}

	return NULL;
}

static struct map *find_map_byentry(struct lookup_context *ctxt, char *entry)
{
	struct map *cur;

	for (cur = ctxt->maps; cur; cur = cur->next) {
		if (!strcmp(cur->mapentry, entry))
			return cur;
	}

	return NULL;
}

static int init_yplib(struct lookup_context *ctxt)
{
	void *dh = NULL;
 	int yplib_len;
	char *yplib_path;
	char *error_string;
	const char *yplib_name = "lookup_yp.so";


	yplib_len = strlen(AUTOFS_LIB_DIR) + strlen(yplib_name);
	yplib_path = alloca(yplib_len + 2);
	if (!yplib_path) {
		crit(MODPREFIX "unable to alloca memory");
		return 0;
	}
	snprintf(yplib_path, yplib_len + 2, "%s/%s",AUTOFS_LIB_DIR,
		 yplib_name);
	if (!(dh = dlopen(yplib_path, RTLD_NOW))) {  
		crit(MODPREFIX "cannot open yp lookup module (%s)",
		     dlerror());
		return 0;
	}

	dlerror(); /* clear any existing error */
	ctxt->lookup_one_included =
		(lookup_one_t)dlsym(dh, "lookup_one_included");
	error_string = dlerror();
	if (error_string) {
		crit(MODPREFIX "dlsym failed for symbol %s, error %s",
		     "lookup_one_included", error_string);
		dlclose(dh);
		return 0;
	}

	ctxt->lookup_readmap_included =
		(lookup_readmap_t)dlsym(dh, "lookup_readmap_included");
	error_string = dlerror();
	if (error_string != NULL) {
		crit(MODPREFIX "dlsym failed for symbol %s, error %s",
		     "lookup_map_included", error_string);
		dlclose(dh);
		ctxt->lookup_one_included = NULL;
		return 0;
	}	

	ctxt->yplib_dh = dh;
	return 1;
}

/*
 *  Takes a string of the form "maptype:mapname" and returns an integer
 *  maptype of either MAPTYPE_YP or MAPTYPE_FILE.  Returns -1 on error.
 */
static int maptype(const char *mapname)
{
	if (!strncmp(mapname, "yp:", 3))
		return MAPTYPE_YP;
	if (!strncmp(mapname, "file:", 5))
		return MAPTYPE_FILE;
	return -1;
}

/*
 *  Return Values:
 *    CHE_OK      -  Success
 *    CHE_FAIL    -  malloc or yp_get_default_domain failed, or there is no
 *                   record of the mapname in the context->maps list, or the
 *                   specified map type is invalid.
 *    CHE_MISSING -  key does not exist in the map
 *    CHE_UPDATED -  either the key was newly added to the cache, or the
 *                   value associated with key changed.
 *
 */
static int lookup_one_included(struct lookup_context *ctxt, char *map,
			       const char *root, const char *key, int key_len)
{
	struct map *mt;

	debug(MODPREFIX "lookup_one_included: looking for map %s, key %s\n",
	      map, key);
	mt = find_map_byentry(ctxt, map);
	if (!mt) {
		debug("find_map_byentry failed\n");
		return CHE_FAIL;
	}

	switch (mt->type) {
	case MAPTYPE_FILE:
		return __lookup_one(mt->name, root, key, key_len, ctxt);

	case MAPTYPE_YP:
		if (!ctxt->lookup_one_included && !init_yplib(ctxt))
			return CHE_FAIL;
		return ctxt->lookup_one_included(map, root, key, key_len);

	default:
		break;
	}

	return CHE_FAIL;
}

/*
 *  This function calls the maptype-specific read_map function.  For file
 *  maps, that simply entails calling __read_map.  For yp maps, we have to
 *  call into the yp module's lookup_map_included function.
 *
 *  Return values:
 *
 *    1 - Success
 *    0 - Failure
 */
static int read_included_map(struct lookup_context *ctxt, char *map,
			     int type, const char *root, time_t age)
{
	struct map *mt;

	debug("read_included_map: %s\n", map);

	if (map == NULL)
		return 0;

	mt = find_map_byname(ctxt, map);
	if (!mt)
		return 0;

	switch (type) {
	case MAPTYPE_FILE:
		return __read_map(map, root, age, ctxt);
	case MAPTYPE_YP:
		if (!ctxt->lookup_readmap_included && !init_yplib(ctxt))
			return 0;

		mt->u.order = map_order(map);
		return ctxt->lookup_readmap_included(map, root, age);
	default:
		error(MODPREFIX "maptype not recognized in map %s.\n", map);
		error(MODPREFIX "valid map types include \"file:\" and \"yp:\"");
		break;
	}
	return 0;
}

static int read_one(FILE *f, char *key, char *mapent)
{
	char *kptr, *p;
	int mapent_len;
	int is_plus_map = 0;
	int ch, nch;
	LOOKUP_STATE state;
	FOUND_STATE getting, gotten;
	ESCAPES escape;

	kptr = key;
	p = NULL;
	mapent_len = 0;
	state = st_begin;
	memset(key, 0, KEY_MAX_LEN + 1);
	memset(mapent, 0, MAPENT_MAX_LEN + 1);
	getting = gotten = got_nothing;
	escape = esc_none;

	/* This is ugly.  We can't just remove \ escape sequences in the value
	   portion of an entry, because the parsing routine may need it. */

	while ((ch = getc(f)) != EOF) {
		switch (escape) {
		case esc_none:
			if (ch == '\\') {
				/* Handle continuation lines */
				if ((nch = getc(f)) == '\n')
					continue;
				ungetc(nch, f);
				escape = esc_char;
			}
			break;

		case esc_char:
			escape = esc_val;
			break;

		case esc_val:
			escape = esc_none;
			break;
		}

		switch (state) {
		case st_begin:
			if (!escape) {
				if (ch == '+')
					is_plus_map = 1;
				if (isspace(ch));
				else if (ch == '#')
					state = st_badent;
				else if (ch == '*') {
					state = st_star;
					*(kptr++) = ch;
				} else if (isprint((char) ch)) {
					state = st_compare;
					*(kptr++) = ch;
				}
			} else if (escape == esc_char);
			else
				state = st_badent;
			break;

		case st_compare:
			if (ch == '\n') {
				if (is_plus_map)
					goto got_it;
				else
					state = st_begin;
			} else if (isspace(ch) && !escape) {
				getting = got_real;
				state = st_entspc;
			} else if (escape == esc_char);
			else if (isprint((char) ch))
				*(kptr++) = ch;
			break;

		case st_star:
			if (ch == '\n')
				state = st_begin;
			else if (isspace(ch) && gotten < got_star && !escape) {
				getting = got_star;
				state = st_entspc;
			} else if (escape != esc_char)
				state = st_badent;
			break;

		case st_badent:
			if (ch == '\n')
				state = st_begin;
			break;

		case st_entspc:
			if (ch == '\n')
				state = st_begin;
			else if (!isspace(ch) || escape) {
				state = st_getent;
				p = mapent;
				gotten = getting;
				*(p++) = ch;
				mapent_len = 1;
			}
			break;

		case st_getent:
			if (ch == '\n') {
				state = st_begin;
				if (gotten == got_real || gotten == getting)
					goto got_it;
			} else if (mapent_len < MAPENT_MAX_LEN) {
				mapent_len++;
				*(p++) = ch;
			}
			break;
		}
		continue;

	      got_it:
		if (is_plus_map)
			return 1;
		if (gotten == got_nothing)
			goto next;

		return 1;

	      next:
		kptr = key;
		p = NULL;
		mapent_len = 0;
		memset(key, 0, KEY_MAX_LEN + 1);
		memset(mapent, 0, MAPENT_MAX_LEN + 1);
		getting = gotten = got_nothing;
		escape = esc_none;
	}

	return 0;
}

/*
 *  Return values:
 *
 *  1 - Success
 *  0 - Failure
 */
static int __read_map(char *map, const char *root,
		      time_t now, struct lookup_context *ctxt)
{
	char key[KEY_MAX_LEN + 1];
	char mapent[MAPENT_MAX_LEN + 1];
	char *mapname;
	FILE *f;
	int  entry;
	time_t age = now ? now : time(NULL);
	struct stat st;
	struct map *mt;

	mt = find_map_byname(ctxt, map);
	if (!mt) {
		crit(MODPREFIX "__read_map: requested map %s not found in"
		     " lookup_context.\n", map);
		return 0;
	}
	
	if (stat(map, &st) < 0) {
		error(MODPREFIX "stat failed on map %s with error %d\n",
		      map, errno);
		return 0;
	}
	mt->u.mtime = st.st_mtime;

	f = fopen(map, "r");
	if (!f) {
		error(MODPREFIX "could not open map file %s", map);
		return 0;
	}

	while(1) {
		entry = read_one(f, key, mapent);
		if (entry && *key == '+') {
			int type;
			char *pathname;

			debug("Looking for plus map: %s, root: %s", key, root);
			mapname = get_nsswitch_map(key+1, mt->name);
			if (!mapname)
				continue;
			type = maptype(mapname);

			pathname = strchr(mapname, ':');
			pathname++;

			if (!add_map(ctxt, type, pathname, key+1)) {
				free(mapname);
				continue;
			}

			if (!read_included_map(ctxt, pathname, type, root, age)) {
				debug("Unable to read included map: %s\n",
				      key+1);
				del_map(ctxt, key+1);
			}
			free(mapname);

		} else if (entry)
			cache_update(root, key, mapent, age);

		if (feof(f))
			break;
	}
	fclose(f);

	/* Clean stale entries from the cache */
	cache_clean(root, age);

	return 1;
}

/*
 *  This is called for the top-level map.  That is to say, the one
 *  specified on the command line.  Included maps go through the __read_map
 *  function.  So, here we can make a valid assumption that the map type
 *  is file.
 */
static int read_map(const char *root, time_t now, struct lookup_context *ctxt)
{
	if (!add_map(ctxt, MAPTYPE_FILE, ctxt->mapname, ctxt->mapname))
		return 0;

	if (!__read_map((char *)ctxt->mapname, root, now, ctxt)) {
		del_map(ctxt, ctxt->mapname);
		return 0;
	}

	return 1;
}

int lookup_ghost(const char *root, int ghost, time_t now, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	struct mapent_cache *me;
	int status = 1;

	/* Clear out the map list.  This is necessary, as otherwise
	 * subsequent calls to lookup_ghost (when re-reading maps, for
	 * example) would result in a failure.
	 */
	map_list_init(ctxt);
	if (!read_map(root, now, ctxt))
		return LKP_FAIL;

	status = cache_ghost(root, ghost, ctxt->mapname, "file", ctxt->parse);

	me = cache_lookup_first();
	/* me NULL => empty map */
	if (me == NULL)
		return LKP_FAIL;

	if (*me->key == '/' && *(root + 1) != '-') {
		me = cache_partial_match(root);
		/* 
		 * me NULL => no entries for this direct mount
		 * root or indirect map
		 */
		if (me == NULL)
			return LKP_FAIL | LKP_INDIRECT;
	}

	return status;
}

static int __lookup_one(const char *map, const char *root, const char *key,
			int key_len, struct lookup_context *ctxt)
{
	char mkey[KEY_MAX_LEN + 1];
	char mapent[MAPENT_MAX_LEN + 1];
	FILE *f;
	int entry;
	int result = 0;
	time_t age = time(NULL);

	debug("__lookup_one: called with map \"%s\".", map);
	f = fopen(map, "r");
	if (!f) {
		error(MODPREFIX "could not open map file %s", map);
		return 0;
	}

	while(1) {
		entry = read_one(f, mkey, mapent);

		if (entry && *mkey == '+') {
			debug("Looking for included map: %s", mkey);
			result = lookup_one_included(ctxt, mkey+1,
						     root, key, key_len);
			if (result == CHE_OK || result == CHE_UPDATED) {
				fclose(f);
				return result;
			}
			debug("Unable to find %s in included map", mkey+1);
		} else if (entry)
			if (strlen(mkey) == key_len &&
			    strncmp(mkey, key, key_len) == 0) {
				fclose(f);
				return cache_update(root, key, mapent, age);
			}

		if (feof(f))
			break;
	}

	fclose(f);

	return CHE_MISSING;
}

static int lookup_one(const char *root,
		      const char *key, int key_len,
		      struct lookup_context *ctxt)
{
	return __lookup_one(ctxt->mapname, root, key, key_len, ctxt);
}

static int lookup_wild(const char *root, struct lookup_context *ctxt)
{
	char mkey[KEY_MAX_LEN + 1];
	char mapent[MAPENT_MAX_LEN + 1];
	FILE *f;
	int entry;
	time_t age = time(NULL);

	f = fopen(ctxt->mapname, "r");
	if (!f) {
		error(MODPREFIX "could not open map file %s", ctxt->mapname);
		return 0;
	}

	while(1) {
		entry = read_one(f, mkey, mapent);
		if (entry)
			if (strncmp(mkey, "*", 1) == 0) {
				fclose(f);
				return cache_update(root, "*", mapent, age);
			}

		if (feof(f))
			break;
	}

	fclose(f);

	return CHE_MISSING;
}

/*
 *  Routine which looks up the yp order (or version number) of the specified
 *  map name.
 *
 *  Returns 0 on failure, or the map order on success.
 */
unsigned int map_order(const char *mapname)
{
	int err;
	char *domainname;
	unsigned int order;

	if ((err = yp_get_default_domain(&domainname)) != YPERR_SUCCESS) {
		error (MODPREFIX "unable to get default yp domain");
		return 0;
	}
	if ((err = yp_order(domainname, mapname, &order)) != YPERR_SUCCESS) {
		error (MODPREFIX "unable to find map, %s in domain, %s",
				mapname, domainname);
		return 0;
	}

	return order;
}

/*
 *  Boolean function which tells whether the map, or any of the included
 *  maps, has changed.
 */
int map_modified(struct lookup_context *ctxt)
{
	struct stat st;
	struct map *mt;
	unsigned int order;

	/* walk through the list of maps to see if they've changed */
	for (mt = ctxt->maps; mt; mt = mt->next) {
		debug("checking for modifications to map %s\n", mt->name);
		switch (mt->type) {
		case MAPTYPE_YP:
			/* test for != instead of >, because an error is
			 * returned as 0 from map_order. */
			order = map_order(mt->name);
			debug("nis map %s order %u, last order %u\n",
			      mt->name, order, mt->u.order);
			if (order != mt->u.order)
				return 1;
			break;

		case MAPTYPE_FILE:
			if (stat(mt->name, &st)) {
				crit(MODPREFIX
				     "file map %s, stat returned %d.",
				     mt->name, errno);
				return 1;
			}
			debug("file map %s mtime %ld, last %ld\n", mt->name,
			      st.st_mtime, mt->u.mtime);
			if (st.st_mtime > mt->u.mtime)
				return 1;
			break;

			/* case MAPTYPE_LDAP: To be implemented*/
		default:
			crit(MODPREFIX "Unrecognized map type %d\n", mt->type);
		}
	}

	debug("map_modified: no changes\n");
	return 0;
}

/*
 *  Returns 0 on success, non-zero on failure.
 */
int lookup_mount(const char *root, const char *name, int name_len, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	char key[KEY_MAX_LEN + 1];
	int key_len;
	char mapent[MAPENT_MAX_LEN + 1];
	struct mapent_cache *me, *exists;
	time_t now = time(NULL);
	time_t t_last_read;
	int ret = 0;

	if (ap.type == LKP_DIRECT)
		key_len = snprintf(key, KEY_MAX_LEN, "%s/%s", root, name);
	else
		key_len = snprintf(key, KEY_MAX_LEN, "%s", name);

	if (key_len > KEY_MAX_LEN)
		return 1;

	me = cache_lookup_first();
	t_last_read = me ? now - me->age : ap.exp_runfreq + 1;

	/* only perform a real lookup if the map was modified */
	if (map_modified(ctxt)) {
		exists = cache_lookup(key);

		ret = lookup_one(root, key, key_len, ctxt);
		if (ret == CHE_FAIL)
			return 1;

		if ((ret & CHE_MISSING) && exists) {
			if (!cache_delete(root, key, CHE_RMPATH))
				rmdir_path(key);

			/* Maybe update wild card map entry */
			if (ap.type == LKP_INDIRECT)
				lookup_wild(root, ctxt);

			/* Have parent update its map */
			if (t_last_read > ap.exp_runfreq)
				kill(getppid(), SIGHUP);
		} else if (ret & CHE_UPDATED) {
			/* Have parent update its map */
			if (t_last_read > ap.exp_runfreq)
				kill(getppid(), SIGHUP);
		}
	}

	me = cache_lookup(key);
	if (me == NULL) {
		/* path component, do submount */
		me = cache_partial_match(key);
		if (me)
			sprintf(mapent, "-fstype=autofs file:%s", ctxt->mapname);
	} else
		sprintf(mapent, me->mapent);

	if (me) {
		debug(MODPREFIX "%s -> %s", key, mapent);
		ret = ctxt->parse->parse_mount(root, name, name_len,
						  mapent, ctxt->parse->context);
	}

	return ret;
}

int lookup_done(void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	int rv = close_parse(ctxt->parse);

	if (ctxt->yplib_dh)
		dlclose(ctxt->yplib_dh);
	free(ctxt);
	cache_release();
	return rv;
}
